#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (C) 2019 Purism SPC
# SPDX-License-Identifier: GPL-3.0+
# Author: David Boddie <david.boddie@puri.sm>

from evdev import AbsInfo, ecodes, InputDevice, InputEvent, UInput
from evdev import list_devices
import time


class DeviceNotFound(Exception):
    pass

class Tester:

    touch_major = 64
    width_major = 64

    def __init__(self, device_file=None, id_=1, debug=False):

        if not device_file:
            device_file = self.find_touch_device()

        self.input = UInput.from_device(InputDevice(device_file))
        self.id_ = id_
        self.debug = debug
        self.touches = {}

    def find_touch_device(self):
        for path in list_devices():
            device = InputDevice(path)
            if ecodes.EV_ABS in device.capabilities():
                return path

        raise DeviceNotFound()

    def new_time(self):
        t = time.time()
        return int(t), int((t - int(t))*1000000)

    def input_event(self, type_, code, value):
        sec, usec = self.new_time()
        return InputEvent(sec, usec, type_, code, value)

    def send_event(self, e):
        if self.debug:
            print(e)
        self.input.write_event(e)

    def syn(self):
        if self.debug:
            print("SYN_REPORT")
        self.input.syn()

    def send_position(self, sec, usec, x, y, touch=None, width=None,
                            button=None, index=None):

        if index != None:
            e = InputEvent(sec, usec, ecodes.EV_ABS, ecodes.ABS_MT_SLOT, index)
            self.send_event(e)

        # Absolute position
        if x != self.touches[index][0]:
            e = InputEvent(sec, usec, ecodes.EV_ABS, ecodes.ABS_MT_POSITION_X, x)
            self.send_event(e)
        if y != self.touches[index][1]:
            e = InputEvent(sec, usec, ecodes.EV_ABS, ecodes.ABS_MT_POSITION_Y, y)
            self.send_event(e)

        # Touch and button information
        if touch != None and width != None and button != None:
            e = InputEvent(sec, usec, ecodes.EV_ABS, ecodes.ABS_MT_TOUCH_MAJOR, touch)
            self.send_event(e)
            e = InputEvent(sec, usec, ecodes.EV_ABS, ecodes.ABS_MT_WIDTH_MAJOR, width)
            self.send_event(e)
            e = InputEvent(sec, usec, ecodes.EV_KEY, ecodes.BTN_TOUCH, button)
            self.send_event(e)

        if not index:
            # Absolute position 
            if x != self.touches[index][0]:
                e = InputEvent(sec, usec, ecodes.EV_ABS, ecodes.ABS_X, x)
                self.send_event(e)
            if y != self.touches[index][1]:
                e = InputEvent(sec, usec, ecodes.EV_ABS, ecodes.ABS_Y, y)
                self.send_event(e)

        self.syn()

        self.touches[index] = (x, y)

    def begin_touch(self, index=None):

        self.touches[index] = (-1, -1)

        sec, usec = self.new_time()
        if index:
            e = InputEvent(sec, usec, ecodes.EV_ABS, ecodes.ABS_MT_SLOT, index)
            self.send_event(e)

        e = InputEvent(sec, usec, ecodes.EV_ABS, ecodes.ABS_MT_TRACKING_ID, self.id_)
        self.send_event(e)
        self.id_ += 1
        if self.id_ > 0x7fffffff:
            self.id_ = 1

        return sec, usec

    def end_touch(self, x=None, y=None, index=None):

        del self.touches[index]

        sec, usec = self.new_time()
        if index != None:
            e = InputEvent(sec, usec, ecodes.EV_ABS, ecodes.ABS_MT_SLOT, index)
            self.send_event(e)

        e = InputEvent(sec, usec, ecodes.EV_ABS, ecodes.ABS_MT_TRACKING_ID, -1)
        self.send_event(e)

        if not self.touches:
            e = InputEvent(sec, usec, ecodes.EV_KEY, ecodes.BTN_TOUCH, 0)
            self.send_event(e)

        self.syn()

    def swipe(self, start, finish, steps, delay):

        sec, usec = self.begin_touch()

        x, y = start
        self.send_position(sec, usec, int(x), int(y), self.touch_major,
                           self.width_major, 1)

        dx, dy = (finish[0] - start[0])/steps, (finish[1] - start[1])/steps
        i = 1
        
        while i < steps:

            x += dx
            y += dy

            time.sleep(delay)
            self.send_position(sec, usec, int(x), int(y))
            sec, usec = self.new_time()

            i += 1

        time.sleep(delay)
        self.send_position(sec, usec, finish[0], finish[1])

        self.end_touch()

    def tap(self, x, y, delay=0.25):

        sec, usec = self.begin_touch()

        self.send_position(sec, usec, int(x), int(y), self.touch_major,
                           self.width_major, 1)
        time.sleep(delay)

        self.end_touch()

    def long_tap(self, x, y):

        self.tap(x, y, delay=2.5)

    def touch(self, index, x, y):

        sec, usec = self.begin_touch(index)
        self.send_position(sec, usec, x, y, self.touch_major, self.width_major,
                           1, index)

    def move_touch(self, index, x, y):

        sec, usec = self.new_time()
        self.send_position(sec, usec, x, y, index=index)

    def pinch(self, lines, steps, delay):

        pos = []
        for i, line in enumerate(lines):
            x, y = line[0]
            self.touch(i, int(x), int(y))
            pos.append([x, y])

        dpos = []
        for line in lines:
            start, finish = line
            dpos.append([(finish[0] - start[0])/steps,
                         (finish[1] - start[1])/steps])

        s = 1

        while s < steps:

            for i, (p, dp) in enumerate(zip(pos, dpos)):
                p[0] += dp[0]
                p[1] += dp[1]

                if p != self.touches[i]:
                    self.move_touch(i, int(p[0]), int(p[1]))

            time.sleep(delay)
            s += 1

        for i, p in enumerate(pos):
            self.move_touch(i, finish[0], finish[1])

        for i in range(2):
            self.end_touch(index=i)

    def send_key(self, key):

        self.input.write(ecodes.EV_KEY, key, 1)
        self.input.write(ecodes.EV_KEY, key, 0)
        self.input.syn()
